@@ -25,6 +25,7 @@ import org.apache.flink.streaming.api.datastream.DataStream;
 import org.apache.flink.streaming.api.datastream.DataStreamSink;
 import org.apache.flink.streaming.api.transformations.PartitionTransformation;
 import org.apache.flink.table.api.EnvironmentSettings;
+import org.apache.flink.table.api.TableException;
 import org.apache.flink.table.api.bridge.java.StreamTableEnvironment;
 import org.apache.flink.table.catalog.ObjectIdentifier;
 import org.apache.flink.table.connector.sink.DataStreamSinkProvider;
@@ -35,7 +36,6 @@ import org.apache.flink.table.factories.DynamicTableFactory;
 import org.apache.flink.table.factories.FactoryUtil;
 import org.apache.flink.table.runtime.connector.sink.SinkRuntimeProviderContext;
 import org.apache.flink.table.store.connector.sink.TableStoreSink;
-import org.apache.flink.table.store.file.FileStoreOptions;
 import org.apache.flink.table.store.file.utils.BlockingIterator;
 import org.apache.flink.table.types.logical.LogicalType;
 import org.apache.flink.table.types.logical.RowType;
@@ -46,6 +46,8 @@ import org.junit.Test;
 
 import javax.annotation.Nullable;
 
+import java.io.File;
+import java.io.IOException;
 import java.util.Arrays;
 import java.util.Collections;
 import java.util.HashMap;
@@ -56,13 +58,14 @@ import java.util.concurrent.TimeUnit;
 
 import static org.apache.flink.table.planner.factories.TestValuesTableFactory.changelogRow;
 import static org.apache.flink.table.planner.factories.TestValuesTableFactory.registerData;
+import static org.apache.flink.table.store.connector.FlinkConnectorOptions.ROOT_PATH;
+import static org.apache.flink.table.store.connector.FlinkConnectorOptions.SCAN_PARALLELISM;
+import static org.apache.flink.table.store.connector.FlinkConnectorOptions.SINK_PARALLELISM;
 import static org.apache.flink.table.store.connector.ReadWriteTableTestUtil.dailyRates;
 import static org.apache.flink.table.store.connector.ReadWriteTableTestUtil.dailyRatesChangelogWithUB;
 import static org.apache.flink.table.store.connector.ReadWriteTableTestUtil.dailyRatesChangelogWithoutUB;
 import static org.apache.flink.table.store.connector.ReadWriteTableTestUtil.rates;
 import static org.apache.flink.table.store.connector.ShowCreateUtil.buildSimpleSelectQuery;
-import static org.apache.flink.table.store.connector.TableStoreFactoryOptions.SCAN_PARALLELISM;
-import static org.apache.flink.table.store.connector.TableStoreFactoryOptions.SINK_PARALLELISM;
 import static org.apache.flink.table.store.connector.TableStoreTestBase.createResolvedTable;
 import static org.assertj.core.api.Assertions.assertThat;
 import static org.assertj.core.api.Assertions.assertThatThrownBy;
@@ -1196,7 +1199,7 @@ public class ReadWriteTableITCase extends ReadWriteTableTestBase {
     }
 
     @Test
-    public void testSinkParallelism() {
+    public void testSinkParallelism() throws IOException {
         testSinkParallelism(null, env.getParallelism());
         testSinkParallelism(23, 23);
     }
@@ -1217,7 +1220,7 @@ public class ReadWriteTableITCase extends ReadWriteTableTestBase {
                         id));
         tEnv.executeSql(
                 String.format(
-                        "create table managed_table with ('path' = '%s') "
+                        "create table managed_table with ('root-path' = '%s') "
                                 + "like dummy_source (excluding options)",
                         rootPath));
         tEnv.executeSql("insert into managed_table select * from dummy_source").await();
@@ -1257,7 +1260,7 @@ public class ReadWriteTableITCase extends ReadWriteTableTestBase {
                                 + "'%s' = '%s'"
                                 + ") like `helper_source_%s` "
                                 + "(excluding options)",
-                        randomSinkId, FileStoreOptions.PATH.key(), rootPath, randomSourceId));
+                        randomSinkId, ROOT_PATH.key(), rootPath, randomSourceId));
 
         // insert multiple times
         tEnv.executeSql(
@@ -1373,7 +1376,7 @@ public class ReadWriteTableITCase extends ReadWriteTableTestBase {
                                 + "'%s' = '%s'"
                                 + ") like `helper_source_%s` "
                                 + "(excluding options)",
-                        randomSinkId, FileStoreOptions.PATH.key(), rootPath, randomSourceId));
+                        randomSinkId, ROOT_PATH.key(), rootPath, randomSourceId));
         tEnv.executeSql(
                         String.format(
                                 "insert into `managed_table_%s` select * from `helper_source_%s`",
@@ -1417,32 +1420,40 @@ public class ReadWriteTableITCase extends ReadWriteTableTestBase {
                 String.format(
                         "CREATE TABLE IF NOT EXISTS rates (\n"
                                 + "currency STRING,\n"
-                                + " rate BIGINT\n"
-                                + ") WITH (\n"
+                                + " rate BIGINT,\n"
+                                + " dt STRING\n"
+                                + ") PARTITIONED BY (dt)\n"
+                                + "WITH (\n"
                                 + " 'bucket' = '2',\n"
-                                + " 'path' = '%s'\n"
+                                + " 'root-path' = '%s'\n"
                                 + ")",
                         rootPath));
-        tEnv.executeSql("INSERT INTO rates VALUES('US Dollar', 102)").await();
+        tEnv.executeSql("INSERT INTO rates VALUES('US Dollar', 102, '2022-06-20')").await();
 
         // increase bucket num from 2 to 3
-        tEnv.executeSql("ALTER TABLE rates SET ('bucket' = '3')");
-        assertThatThrownBy(
-                        () -> tEnv.executeSql("INSERT INTO rates VALUES('US Dollar', 102)").await())
-                .hasRootCauseInstanceOf(IllegalStateException.class)
-                .hasRootCauseMessage(
-                        "Bucket number has been changed. Manifest might be corrupted.");
-        assertThatThrownBy(() -> tEnv.executeSql("SELECT * FROM rates").await())
-                .hasRootCauseInstanceOf(IllegalStateException.class)
-                .hasRootCauseMessage(
-                        "Bucket number has been changed. Manifest might be corrupted.");
+        assertChangeBucketWithoutRescale(3);
 
         // decrease bucket num from 3 to 1
-        tEnv.executeSql("ALTER TABLE rates RESET ('bucket')");
-        assertThatThrownBy(() -> tEnv.executeSql("SELECT * FROM rates").await())
-                .hasRootCauseInstanceOf(IllegalStateException.class)
-                .hasRootCauseMessage(
-                        "Bucket number has been changed. Manifest might be corrupted.");
+        assertChangeBucketWithoutRescale(1);
+    }
+
+    private void assertChangeBucketWithoutRescale(int bucketNum) throws Exception {
+        tEnv.executeSql(String.format("ALTER TABLE rates SET ('bucket' = '%d')", bucketNum));
+        // read is ok
+        assertThat(BlockingIterator.of(tEnv.executeSql("SELECT * FROM rates").collect()).collect())
+                .containsExactlyInAnyOrder(changelogRow("+I", "US Dollar", 102L, "2022-06-20"));
+        assertThatThrownBy(
+                        () ->
+                                tEnv.executeSql(
+                                                "INSERT INTO rates VALUES('US Dollar', 102, '2022-06-20')")
+                                        .await())
+                .getRootCause()
+                .isInstanceOf(TableException.class)
+                .hasMessage(
+                        String.format(
+                                "Try to write partition {dt=2022-06-20} with a new bucket num %d, but the previous bucket num is 2. "
+                                        + "Please switch to batch mode, and perform INSERT OVERWRITE to rescale current data layout first.",
+                                bucketNum));
     }
 
     @Test
@@ -1573,12 +1584,18 @@ public class ReadWriteTableITCase extends ReadWriteTableTestBase {
         return stream.getParallelism();
     }
 
-    private void testSinkParallelism(Integer configParallelism, int expectedParallelism) {
+    private void testSinkParallelism(Integer configParallelism, int expectedParallelism)
+            throws IOException {
         // 1. create a mock table sink
         Map<String, String> options = new HashMap<>();
         if (configParallelism != null) {
             options.put(SINK_PARALLELISM.key(), configParallelism.toString());
         }
+        options.put(
+                "path",
+                new File(TEMPORARY_FOLDER.newFolder(), UUID.randomUUID().toString())
+                        .toURI()
+                        .toString());
 
         DynamicTableFactory.Context context =
                 new FactoryUtil.DefaultDynamicTableContext(
@@ -1594,7 +1611,8 @@ public class ReadWriteTableITCase extends ReadWriteTableTestBase {
                         new Configuration(),
                         Thread.currentThread().getContextClassLoader(),
                         false);
-        DynamicTableSink tableSink = new TableStoreFactory().createDynamicTableSink(context);
+        new TableStoreManagedFactory().onCreateTable(context, false);
+        DynamicTableSink tableSink = new TableStoreManagedFactory().createDynamicTableSink(context);
         assertThat(tableSink).isInstanceOf(TableStoreSink.class);
 
         // 2. get sink provider
