@@ -21,21 +21,22 @@ package org.apache.flink.table.store.file.operation;
 import org.apache.flink.configuration.MemorySize;
 import org.apache.flink.core.fs.FileSystem;
 import org.apache.flink.core.fs.Path;
+import org.apache.flink.table.data.RowData;
 import org.apache.flink.table.data.binary.BinaryRowData;
 import org.apache.flink.table.store.file.Snapshot;
-import org.apache.flink.table.store.file.ValueKind;
+import org.apache.flink.table.store.file.data.DataFileMeta;
+import org.apache.flink.table.store.file.manifest.FileKind;
 import org.apache.flink.table.store.file.manifest.ManifestCommittable;
 import org.apache.flink.table.store.file.manifest.ManifestEntry;
 import org.apache.flink.table.store.file.manifest.ManifestFile;
 import org.apache.flink.table.store.file.manifest.ManifestFileMeta;
 import org.apache.flink.table.store.file.manifest.ManifestList;
-import org.apache.flink.table.store.file.mergetree.sst.SstFileMeta;
 import org.apache.flink.table.store.file.predicate.Predicate;
 import org.apache.flink.table.store.file.predicate.PredicateConverter;
+import org.apache.flink.table.store.file.utils.AtomicFileWriter;
 import org.apache.flink.table.store.file.utils.FileStorePathFactory;
-import org.apache.flink.table.store.file.utils.FileUtils;
-import org.apache.flink.table.store.file.utils.RowDataToObjectArrayConverter;
-import org.apache.flink.table.store.file.utils.SnapshotFinder;
+import org.apache.flink.table.store.file.utils.SnapshotManager;
+import org.apache.flink.table.store.utils.RowDataToObjectArrayConverter;
 import org.apache.flink.table.types.logical.RowType;
 
 import org.slf4j.Logger;
@@ -43,12 +44,15 @@ import org.slf4j.LoggerFactory;
 
 import javax.annotation.Nullable;
 
-import java.io.IOException;
 import java.util.ArrayList;
+import java.util.Collection;
+import java.util.Comparator;
+import java.util.HashMap;
 import java.util.HashSet;
 import java.util.LinkedHashMap;
 import java.util.List;
 import java.util.Map;
+import java.util.Objects;
 import java.util.Set;
 import java.util.concurrent.Callable;
 import java.util.stream.Collectors;
@@ -73,41 +77,52 @@ public class FileStoreCommitImpl implements FileStoreCommit {
 
     private static final Logger LOG = LoggerFactory.getLogger(FileStoreCommitImpl.class);
 
+    private final long schemaId;
     private final String commitUser;
     private final RowType partitionType;
     private final RowDataToObjectArrayConverter partitionObjectConverter;
     private final FileStorePathFactory pathFactory;
+    private final SnapshotManager snapshotManager;
     private final ManifestFile manifestFile;
     private final ManifestList manifestList;
     private final FileStoreScan scan;
     private final int numBucket;
     private final MemorySize manifestTargetSize;
     private final int manifestMergeMinCount;
+    @Nullable private final Comparator<RowData> keyComparator;
 
     @Nullable private Lock lock;
+    private boolean createEmptyCommit;
 
     public FileStoreCommitImpl(
+            long schemaId,
             String commitUser,
             RowType partitionType,
             FileStorePathFactory pathFactory,
+            SnapshotManager snapshotManager,
             ManifestFile.Factory manifestFileFactory,
             ManifestList.Factory manifestListFactory,
             FileStoreScan scan,
             int numBucket,
             MemorySize manifestTargetSize,
-            int manifestMergeMinCount) {
+            int manifestMergeMinCount,
+            @Nullable Comparator<RowData> keyComparator) {
+        this.schemaId = schemaId;
         this.commitUser = commitUser;
         this.partitionType = partitionType;
         this.partitionObjectConverter = new RowDataToObjectArrayConverter(partitionType);
         this.pathFactory = pathFactory;
+        this.snapshotManager = snapshotManager;
         this.manifestFile = manifestFileFactory.create();
         this.manifestList = manifestListFactory.create();
         this.scan = scan;
         this.numBucket = numBucket;
         this.manifestTargetSize = manifestTargetSize;
         this.manifestMergeMinCount = manifestMergeMinCount;
+        this.keyComparator = keyComparator;
 
         this.lock = null;
+        this.createEmptyCommit = false;
     }
 
     @Override
@@ -116,6 +131,12 @@ public class FileStoreCommitImpl implements FileStoreCommit {
         return this;
     }
 
+    @Override
+    public FileStoreCommit withCreateEmptyCommit(boolean createEmptyCommit) {
+        this.createEmptyCommit = createEmptyCommit;
+        return this;
+    }
+
     @Override
     public List<ManifestCommittable> filterCommitted(List<ManifestCommittable> committableList) {
         // nothing to filter, fast exit
@@ -124,7 +145,7 @@ public class FileStoreCommitImpl implements FileStoreCommit {
         }
 
         // if there is no previous snapshots then nothing should be filtered
-        Long latestSnapshotId = pathFactory.latestSnapshotId();
+        Long latestSnapshotId = snapshotManager.latestSnapshotId();
         if (latestSnapshotId == null) {
             return committableList;
         }
@@ -136,16 +157,11 @@ public class FileStoreCommitImpl implements FileStoreCommit {
         }
 
         for (long id = latestSnapshotId; id >= Snapshot.FIRST_SNAPSHOT_ID; id--) {
-            Path snapshotPath = pathFactory.toSnapshotPath(id);
-            try {
-                if (!snapshotPath.getFileSystem().exists(snapshotPath)) {
-                    // snapshots before this are expired
-                    break;
-                }
-            } catch (IOException e) {
-                throw new RuntimeException("Cannot determine if snapshot #" + id + " exists.", e);
+            if (!snapshotManager.snapshotExists(id)) {
+                // snapshots before this are expired
+                break;
             }
-            Snapshot snapshot = Snapshot.fromPath(snapshotPath);
+            Snapshot snapshot = snapshotManager.snapshot(id);
             if (commitUser.equals(snapshot.commitUser())) {
                 if (identifiers.containsKey(snapshot.commitIdentifier())) {
                     identifiers.remove(snapshot.commitIdentifier());
@@ -166,17 +182,19 @@ public class FileStoreCommitImpl implements FileStoreCommit {
             LOG.debug("Ready to commit\n" + committable.toString());
         }
 
-        List<ManifestEntry> appendChanges = collectChanges(committable.newFiles(), ValueKind.ADD);
-        tryCommit(
-                appendChanges,
-                committable.identifier(),
-                committable.logOffsets(),
-                Snapshot.CommitKind.APPEND,
-                false);
+        List<ManifestEntry> appendChanges = collectChanges(committable.newFiles(), FileKind.ADD);
+        if (createEmptyCommit || !appendChanges.isEmpty()) {
+            tryCommit(
+                    appendChanges,
+                    committable.identifier(),
+                    committable.logOffsets(),
+                    Snapshot.CommitKind.APPEND,
+                    false);
+        }
 
         List<ManifestEntry> compactChanges = new ArrayList<>();
-        compactChanges.addAll(collectChanges(committable.compactBefore(), ValueKind.DELETE));
-        compactChanges.addAll(collectChanges(committable.compactAfter(), ValueKind.ADD));
+        compactChanges.addAll(collectChanges(committable.compactBefore(), FileKind.DELETE));
+        compactChanges.addAll(collectChanges(committable.compactAfter(), FileKind.ADD));
         if (!compactChanges.isEmpty()) {
             tryCommit(
                     compactChanges,
@@ -200,9 +218,9 @@ public class FileStoreCommitImpl implements FileStoreCommit {
                             + committable.toString());
         }
 
-        List<ManifestEntry> appendChanges = collectChanges(committable.newFiles(), ValueKind.ADD);
+        List<ManifestEntry> appendChanges = collectChanges(committable.newFiles(), FileKind.ADD);
         // sanity check, all changes must be done within the given partition
-        Predicate partitionFilter = PredicateConverter.CONVERTER.fromMap(partition, partitionType);
+        Predicate partitionFilter = PredicateConverter.fromMap(partition, partitionType);
         if (partitionFilter != null) {
             for (ManifestEntry entry : appendChanges) {
                 if (!partitionFilter.test(partitionObjectConverter.convert(entry.partition()))) {
@@ -220,8 +238,8 @@ public class FileStoreCommitImpl implements FileStoreCommit {
                 partitionFilter, appendChanges, committable.identifier(), committable.logOffsets());
 
         List<ManifestEntry> compactChanges = new ArrayList<>();
-        compactChanges.addAll(collectChanges(committable.compactBefore(), ValueKind.DELETE));
-        compactChanges.addAll(collectChanges(committable.compactAfter(), ValueKind.ADD));
+        compactChanges.addAll(collectChanges(committable.compactBefore(), FileKind.DELETE));
+        compactChanges.addAll(collectChanges(committable.compactAfter(), FileKind.ADD));
         if (!compactChanges.isEmpty()) {
             tryCommit(
                     compactChanges,
@@ -237,11 +255,11 @@ public class FileStoreCommitImpl implements FileStoreCommit {
             String hash,
             Map<Integer, Long> logOffsets,
             Snapshot.CommitKind commitKind,
-            boolean checkDeletedFiles) {
+            boolean checkFileConflicts) {
         while (true) {
-            Long latestSnapshotId = pathFactory.latestSnapshotId();
+            Long latestSnapshotId = snapshotManager.latestSnapshotId();
             if (tryCommitOnce(
-                    changes, hash, logOffsets, commitKind, latestSnapshotId, checkDeletedFiles)) {
+                    changes, hash, logOffsets, commitKind, latestSnapshotId, checkFileConflicts)) {
                 break;
             }
         }
@@ -253,7 +271,7 @@ public class FileStoreCommitImpl implements FileStoreCommit {
             String identifier,
             Map<Integer, Long> logOffsets) {
         while (true) {
-            Long latestSnapshotId = pathFactory.latestSnapshotId();
+            Long latestSnapshotId = snapshotManager.latestSnapshotId();
 
             List<ManifestEntry> changesWithOverwrite = new ArrayList<>();
             if (latestSnapshotId != null) {
@@ -265,7 +283,7 @@ public class FileStoreCommitImpl implements FileStoreCommit {
                 for (ManifestEntry entry : currentEntries) {
                     changesWithOverwrite.add(
                             new ManifestEntry(
-                                    ValueKind.DELETE,
+                                    FileKind.DELETE,
                                     entry.partition(),
                                     entry.bucket(),
                                     entry.totalBuckets(),
@@ -287,11 +305,11 @@ public class FileStoreCommitImpl implements FileStoreCommit {
     }
 
     private List<ManifestEntry> collectChanges(
-            Map<BinaryRowData, Map<Integer, List<SstFileMeta>>> map, ValueKind kind) {
+            Map<BinaryRowData, Map<Integer, List<DataFileMeta>>> map, FileKind kind) {
         List<ManifestEntry> changes = new ArrayList<>();
-        for (Map.Entry<BinaryRowData, Map<Integer, List<SstFileMeta>>> entryWithPartition :
+        for (Map.Entry<BinaryRowData, Map<Integer, List<DataFileMeta>>> entryWithPartition :
                 map.entrySet()) {
-            for (Map.Entry<Integer, List<SstFileMeta>> entryWithBucket :
+            for (Map.Entry<Integer, List<DataFileMeta>> entryWithBucket :
                     entryWithPartition.getValue().entrySet()) {
                 changes.addAll(
                         entryWithBucket.getValue().stream()
@@ -315,11 +333,10 @@ public class FileStoreCommitImpl implements FileStoreCommit {
             Map<Integer, Long> logOffsets,
             Snapshot.CommitKind commitKind,
             Long latestSnapshotId,
-            boolean checkDeletedFiles) {
+            boolean checkFileConflicts) {
         long newSnapshotId =
                 latestSnapshotId == null ? Snapshot.FIRST_SNAPSHOT_ID : latestSnapshotId + 1;
-        Path newSnapshotPath = pathFactory.toSnapshotPath(newSnapshotId);
-        Path tmpSnapshotPath = pathFactory.toTmpSnapshotPath(newSnapshotId);
+        Path newSnapshotPath = snapshotManager.snapshotPath(newSnapshotId);
 
         if (LOG.isDebugEnabled()) {
             LOG.debug("Ready to commit changes to snapshot #" + newSnapshotId);
@@ -330,10 +347,10 @@ public class FileStoreCommitImpl implements FileStoreCommit {
 
         Snapshot latestSnapshot = null;
         if (latestSnapshotId != null) {
-            if (checkDeletedFiles) {
+            if (checkFileConflicts) {
                 noConflictsOrFail(latestSnapshotId, changes);
             }
-            latestSnapshot = Snapshot.fromPath(pathFactory.toSnapshotPath(latestSnapshotId));
+            latestSnapshot = snapshotManager.snapshot(latestSnapshotId);
         }
 
         Snapshot newSnapshot;
@@ -367,6 +384,7 @@ public class FileStoreCommitImpl implements FileStoreCommit {
             newSnapshot =
                     new Snapshot(
                             newSnapshotId,
+                            schemaId,
                             previousChangesListName,
                             newChangesListName,
                             commitUser,
@@ -374,15 +392,9 @@ public class FileStoreCommitImpl implements FileStoreCommit {
                             commitKind,
                             System.currentTimeMillis(),
                             logOffsets);
-            FileUtils.writeFileUtf8(tmpSnapshotPath, newSnapshot.toJson());
         } catch (Throwable e) {
             // fails when preparing for commit, we should clean up
-            cleanUpTmpSnapshot(
-                    tmpSnapshotPath,
-                    previousChangesListName,
-                    newChangesListName,
-                    oldMetas,
-                    newMetas);
+            cleanUpTmpManifests(previousChangesListName, newChangesListName, oldMetas, newMetas);
             throw new RuntimeException(
                     String.format(
                             "Exception occurs when preparing snapshot #%d (path %s) by user %s "
@@ -397,15 +409,18 @@ public class FileStoreCommitImpl implements FileStoreCommit {
 
         boolean success;
         try {
-            FileSystem fs = tmpSnapshotPath.getFileSystem();
-            // atomic rename
-            // TODO rename is not work for object store, use recoverable writer
+            FileSystem fs = newSnapshotPath.getFileSystem();
             Callable<Boolean> callable =
                     () -> {
-                        boolean committed = fs.rename(tmpSnapshotPath, newSnapshotPath);
+                        if (fs.exists(newSnapshotPath)) {
+                            return false;
+                        }
+
+                        boolean committed =
+                                AtomicFileWriter.writeFileUtf8(
+                                        newSnapshotPath, newSnapshot.toJson());
                         if (committed) {
-                            SnapshotFinder.commitLatestHint(
-                                    pathFactory.snapshotDirectory(), newSnapshotId);
+                            snapshotManager.commitLatestHint(newSnapshotId);
                         }
                         return committed;
                     };
@@ -431,7 +446,7 @@ public class FileStoreCommitImpl implements FileStoreCommit {
                                     + "with identifier %s and kind %s. "
                                     + "Cannot clean up because we can't determine the success.",
                             newSnapshotId,
-                            newSnapshotPath.toString(),
+                            newSnapshotPath,
                             commitUser,
                             identifier,
                             commitKind.name()),
@@ -445,7 +460,7 @@ public class FileStoreCommitImpl implements FileStoreCommit {
                                 "Successfully commit snapshot #%d (path %s) by user %s "
                                         + "with identifier %s and kind %s.",
                                 newSnapshotId,
-                                newSnapshotPath.toString(),
+                                newSnapshotPath,
                                 commitUser,
                                 identifier,
                                 commitKind.name()));
@@ -456,72 +471,81 @@ public class FileStoreCommitImpl implements FileStoreCommit {
         // atomic rename fails, clean up and try again
         LOG.warn(
                 String.format(
-                        "Atomic rename failed for snapshot #%d (path %s) by user %s "
+                        "Atomic commit failed for snapshot #%d (path %s) by user %s "
                                 + "with identifier %s and kind %s. "
                                 + "Clean up and try again.",
-                        newSnapshotId,
-                        newSnapshotPath.toString(),
-                        commitUser,
-                        identifier,
-                        commitKind.name()));
-        cleanUpTmpSnapshot(
-                tmpSnapshotPath, previousChangesListName, newChangesListName, oldMetas, newMetas);
+                        newSnapshotId, newSnapshotPath, commitUser, identifier, commitKind.name()));
+        cleanUpTmpManifests(previousChangesListName, newChangesListName, oldMetas, newMetas);
         return false;
     }
 
     private void noConflictsOrFail(long snapshotId, List<ManifestEntry> changes) {
-        Set<ManifestEntry.Identifier> removedFiles =
-                changes.stream()
-                        .filter(e -> e.kind().equals(ValueKind.DELETE))
-                        .map(ManifestEntry::identifier)
-                        .collect(Collectors.toSet());
-        if (removedFiles.isEmpty()) {
-            // early exit for append only changes
-            return;
-        }
-
         List<BinaryRowData> changedPartitions =
                 changes.stream()
                         .map(ManifestEntry::partition)
                         .distinct()
                         .collect(Collectors.toList());
+        List<ManifestEntry> allEntries;
         try {
-            for (ManifestEntry entry :
-                    scan.withSnapshot(snapshotId)
-                            .withPartitionFilter(changedPartitions)
-                            .plan()
-                            .files()) {
-                removedFiles.remove(entry.identifier());
-            }
+            allEntries =
+                    new ArrayList<>(
+                            scan.withSnapshot(snapshotId)
+                                    .withPartitionFilter(changedPartitions)
+                                    .plan()
+                                    .files());
         } catch (Throwable e) {
             throw new RuntimeException("Cannot determine if conflicts exist.", e);
         }
+        allEntries.addAll(changes);
 
-        if (!removedFiles.isEmpty()) {
+        Collection<ManifestEntry> mergedEntries;
+        try {
+            // merge manifest entries and also check if the files we want to delete are still there
+            mergedEntries = ManifestEntry.mergeManifestEntries(allEntries);
+        } catch (Throwable e) {
             throw new RuntimeException(
-                    "Conflicts detected on:\n"
-                            + removedFiles.stream()
-                                    .map(
-                                            i ->
-                                                    pathFactory.getPartitionString(i.partition)
-                                                            + ", bucket "
-                                                            + i.bucket
-                                                            + ", level "
-                                                            + i.level
-                                                            + ", file "
-                                                            + i.fileName)
-                                    .collect(Collectors.joining("\n")));
+                    "File deletion conflicts detected! Give up committing compact changes.", e);
+        }
+
+        // fast exit for file store without keys
+        if (keyComparator == null) {
+            return;
+        }
+
+        // group entries by partitions, buckets and levels
+        Map<LevelIdentifier, List<ManifestEntry>> levels = new HashMap<>();
+        for (ManifestEntry entry : mergedEntries) {
+            int level = entry.file().level();
+            if (level >= 1) {
+                levels.computeIfAbsent(
+                                new LevelIdentifier(entry.partition(), entry.bucket(), level),
+                                lv -> new ArrayList<>())
+                        .add(entry);
+            }
+        }
+
+        // check for all LSM level >= 1, key ranges of files do not intersect
+        for (List<ManifestEntry> entries : levels.values()) {
+            entries.sort((a, b) -> keyComparator.compare(a.file().minKey(), b.file().minKey()));
+            for (int i = 0; i + 1 < entries.size(); i++) {
+                ManifestEntry a = entries.get(i);
+                ManifestEntry b = entries.get(i + 1);
+                if (keyComparator.compare(a.file().maxKey(), b.file().minKey()) >= 0) {
+                    throw new RuntimeException(
+                            "LSM conflicts detected! Give up committing compact changes. Conflict files are:\n"
+                                    + a.identifier().toString(pathFactory)
+                                    + "\n"
+                                    + b.identifier().toString(pathFactory));
+                }
+            }
         }
     }
 
-    private void cleanUpTmpSnapshot(
-            Path tmpSnapshotPath,
+    private void cleanUpTmpManifests(
             String previousChangesListName,
             String newChangesListName,
             List<ManifestFileMeta> oldMetas,
             List<ManifestFileMeta> newMetas) {
-        // clean up tmp snapshot file
-        FileUtils.deleteOrWarn(tmpSnapshotPath);
         // clean up newly created manifest list
         if (previousChangesListName != null) {
             manifestList.delete(previousChangesListName);
@@ -537,4 +561,33 @@ public class FileStoreCommitImpl implements FileStoreCommit {
             }
         }
     }
+
+    private static class LevelIdentifier {
+
+        private final BinaryRowData partition;
+        private final int bucket;
+        private final int level;
+
+        private LevelIdentifier(BinaryRowData partition, int bucket, int level) {
+            this.partition = partition;
+            this.bucket = bucket;
+            this.level = level;
+        }
+
+        @Override
+        public boolean equals(Object o) {
+            if (!(o instanceof LevelIdentifier)) {
+                return false;
+            }
+            LevelIdentifier that = (LevelIdentifier) o;
+            return Objects.equals(partition, that.partition)
+                    && bucket == that.bucket
+                    && level == that.level;
+        }
+
+        @Override
+        public int hashCode() {
+            return Objects.hash(partition, bucket, level);
+        }
+    }
 }
