@@ -21,28 +21,30 @@ package org.apache.flink.table.store.file;
 import org.apache.flink.configuration.Configuration;
 import org.apache.flink.configuration.MemorySize;
 import org.apache.flink.core.fs.Path;
-import org.apache.flink.table.catalog.ObjectIdentifier;
 import org.apache.flink.table.data.binary.BinaryRowData;
 import org.apache.flink.table.runtime.typeutils.RowDataSerializer;
+import org.apache.flink.table.store.CoreOptions;
+import org.apache.flink.table.store.file.data.DataFileMeta;
 import org.apache.flink.table.store.file.manifest.ManifestCommittable;
 import org.apache.flink.table.store.file.manifest.ManifestEntry;
 import org.apache.flink.table.store.file.manifest.ManifestFileMeta;
 import org.apache.flink.table.store.file.manifest.ManifestList;
+import org.apache.flink.table.store.file.memory.HeapMemorySegmentPool;
+import org.apache.flink.table.store.file.memory.MemoryOwner;
 import org.apache.flink.table.store.file.mergetree.Increment;
-import org.apache.flink.table.store.file.mergetree.MergeTreeOptions;
 import org.apache.flink.table.store.file.mergetree.compact.MergeFunction;
-import org.apache.flink.table.store.file.mergetree.sst.SstFileMeta;
 import org.apache.flink.table.store.file.operation.FileStoreCommit;
 import org.apache.flink.table.store.file.operation.FileStoreExpireImpl;
 import org.apache.flink.table.store.file.operation.FileStoreRead;
 import org.apache.flink.table.store.file.operation.FileStoreScan;
 import org.apache.flink.table.store.file.operation.FileStoreWrite;
+import org.apache.flink.table.store.file.schema.SchemaManager;
 import org.apache.flink.table.store.file.utils.FileStorePathFactory;
 import org.apache.flink.table.store.file.utils.RecordReaderIterator;
-import org.apache.flink.table.store.file.utils.RecordWriter;
-import org.apache.flink.table.store.file.utils.SnapshotFinder;
+import org.apache.flink.table.store.file.utils.SnapshotManager;
+import org.apache.flink.table.store.file.writer.RecordWriter;
+import org.apache.flink.table.store.table.source.Split;
 import org.apache.flink.table.types.logical.RowType;
-import org.apache.flink.util.function.QuadFunction;
 
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
@@ -69,13 +71,18 @@ import java.util.stream.Collectors;
 import static org.assertj.core.api.Assertions.assertThat;
 
 /** {@link FileStore} for tests. */
-public class TestFileStore extends FileStoreImpl {
+public class TestFileStore extends KeyValueFileStore {
 
     private static final Logger LOG = LoggerFactory.getLogger(TestFileStore.class);
 
+    public static final MemorySize WRITE_BUFFER_SIZE = MemorySize.parse("16 kb");
+
+    public static final MemorySize PAGE_SIZE = MemorySize.parse("4 kb");
+
     private final String root;
     private final RowDataSerializer keySerializer;
     private final RowDataSerializer valueSerializer;
+    private final String user;
 
     public static TestFileStore create(
             String format,
@@ -87,39 +94,43 @@ public class TestFileStore extends FileStoreImpl {
             MergeFunction mergeFunction) {
         Configuration conf = new Configuration();
 
-        conf.set(MergeTreeOptions.WRITE_BUFFER_SIZE, MemorySize.parse("16 kb"));
-        conf.set(MergeTreeOptions.PAGE_SIZE, MemorySize.parse("4 kb"));
-        conf.set(MergeTreeOptions.TARGET_FILE_SIZE, MemorySize.parse("1 kb"));
+        conf.set(CoreOptions.WRITE_BUFFER_SIZE, WRITE_BUFFER_SIZE);
+        conf.set(CoreOptions.PAGE_SIZE, PAGE_SIZE);
+        conf.set(CoreOptions.TARGET_FILE_SIZE, MemorySize.parse("1 kb"));
 
         conf.set(
-                FileStoreOptions.MANIFEST_TARGET_FILE_SIZE,
+                CoreOptions.MANIFEST_TARGET_FILE_SIZE,
                 MemorySize.parse((ThreadLocalRandom.current().nextInt(16) + 1) + "kb"));
 
-        conf.set(FileStoreOptions.FILE_FORMAT, format);
-        conf.set(FileStoreOptions.MANIFEST_FORMAT, format);
-        conf.set(FileStoreOptions.PATH, root);
-        conf.set(FileStoreOptions.BUCKET, numBuckets);
+        conf.set(CoreOptions.FILE_FORMAT, format);
+        conf.set(CoreOptions.MANIFEST_FORMAT, format);
+        conf.set(CoreOptions.PATH, root);
+        conf.set(CoreOptions.BUCKET, numBuckets);
 
-        return new TestFileStore(conf, partitionType, keyType, valueType, mergeFunction);
+        return new TestFileStore(
+                root, new CoreOptions(conf), partitionType, keyType, valueType, mergeFunction);
     }
 
-    public TestFileStore(
-            Configuration conf,
+    private TestFileStore(
+            String root,
+            CoreOptions options,
             RowType partitionType,
             RowType keyType,
             RowType valueType,
             MergeFunction mergeFunction) {
         super(
-                ObjectIdentifier.of("catalog", "database", "table"),
-                conf,
-                UUID.randomUUID().toString(),
+                new SchemaManager(options.path()),
+                0L,
+                options,
                 partitionType,
                 keyType,
+                keyType,
                 valueType,
                 mergeFunction);
-        this.root = conf.getString(FileStoreOptions.PATH);
+        this.root = root;
         this.keySerializer = new RowDataSerializer(keyType);
         this.valueSerializer = new RowDataSerializer(valueType);
+        this.user = UUID.randomUUID().toString();
     }
 
     public FileStoreExpireImpl newExpire(
@@ -129,6 +140,7 @@ public class TestFileStore extends FileStoreImpl {
                 numRetainedMax,
                 millisRetained,
                 pathFactory(),
+                snapshotManager(),
                 manifestFileFactory(),
                 manifestListFactory());
     }
@@ -151,7 +163,7 @@ public class TestFileStore extends FileStoreImpl {
                 kvs,
                 partitionCalculator,
                 bucketCalculator,
-                FileStoreWrite::createWriter,
+                false,
                 (commit, committable) -> {
                     logOffsets.forEach(committable::addLogOffset);
                     commit.commit(committable, Collections.emptyMap());
@@ -168,59 +180,68 @@ public class TestFileStore extends FileStoreImpl {
                 kvs,
                 partitionCalculator,
                 bucketCalculator,
-                FileStoreWrite::createEmptyWriter,
+                true,
                 (commit, committable) ->
                         commit.overwrite(partition, committable, Collections.emptyMap()));
     }
 
-    private List<Snapshot> commitDataImpl(
+    public List<Snapshot> commitDataImpl(
             List<KeyValue> kvs,
             Function<KeyValue, BinaryRowData> partitionCalculator,
             Function<KeyValue, Integer> bucketCalculator,
-            QuadFunction<FileStoreWrite, BinaryRowData, Integer, ExecutorService, RecordWriter>
-                    createWriterFunction,
+            boolean emptyWriter,
             BiConsumer<FileStoreCommit, ManifestCommittable> commitFunction)
             throws Exception {
-        FileStoreWrite write = newWrite();
-        Map<BinaryRowData, Map<Integer, RecordWriter>> writers = new HashMap<>();
+        FileStoreWrite<KeyValue> write = newWrite();
+        Map<BinaryRowData, Map<Integer, RecordWriter<KeyValue>>> writers = new HashMap<>();
         for (KeyValue kv : kvs) {
             BinaryRowData partition = partitionCalculator.apply(kv);
             int bucket = bucketCalculator.apply(kv);
-            writers.compute(partition, (p, m) -> m == null ? new HashMap<>() : m)
+            writers.computeIfAbsent(partition, p -> new HashMap<>())
                     .compute(
                             bucket,
                             (b, w) -> {
                                 if (w == null) {
                                     ExecutorService service = Executors.newSingleThreadExecutor();
-                                    return createWriterFunction.apply(
-                                            write, partition, bucket, service);
+                                    RecordWriter<KeyValue> writer =
+                                            emptyWriter
+                                                    ? write.createEmptyWriter(
+                                                            partition, bucket, service)
+                                                    : write.createWriter(
+                                                            partition, bucket, service);
+                                    ((MemoryOwner) writer)
+                                            .setMemoryPool(
+                                                    new HeapMemorySegmentPool(
+                                                            WRITE_BUFFER_SIZE.getBytes(),
+                                                            (int) PAGE_SIZE.getBytes()));
+                                    return writer;
                                 } else {
                                     return w;
                                 }
                             })
-                    .write(kv.valueKind(), kv.key(), kv.value());
+                    .write(kv);
         }
 
-        FileStoreCommit commit = newCommit();
+        FileStoreCommit commit = newCommit(user);
         ManifestCommittable committable =
                 new ManifestCommittable(String.valueOf(new Random().nextLong()));
-        for (Map.Entry<BinaryRowData, Map<Integer, RecordWriter>> entryWithPartition :
+        for (Map.Entry<BinaryRowData, Map<Integer, RecordWriter<KeyValue>>> entryWithPartition :
                 writers.entrySet()) {
-            for (Map.Entry<Integer, RecordWriter> entryWithBucket :
+            for (Map.Entry<Integer, RecordWriter<KeyValue>> entryWithBucket :
                     entryWithPartition.getValue().entrySet()) {
-                Increment increment = entryWithBucket.getValue().prepareCommit();
+                Increment increment = entryWithBucket.getValue().prepareCommit(emptyWriter);
                 committable.addFileCommittable(
                         entryWithPartition.getKey(), entryWithBucket.getKey(), increment);
             }
         }
 
-        FileStorePathFactory pathFactory = pathFactory();
-        Long snapshotIdBeforeCommit = pathFactory.latestSnapshotId();
+        SnapshotManager snapshotManager = snapshotManager();
+        Long snapshotIdBeforeCommit = snapshotManager.latestSnapshotId();
         if (snapshotIdBeforeCommit == null) {
             snapshotIdBeforeCommit = Snapshot.FIRST_SNAPSHOT_ID - 1;
         }
         commitFunction.accept(commit, committable);
-        Long snapshotIdAfterCommit = pathFactory.latestSnapshotId();
+        Long snapshotIdAfterCommit = snapshotManager.latestSnapshotId();
         if (snapshotIdAfterCommit == null) {
             snapshotIdAfterCommit = Snapshot.FIRST_SNAPSHOT_ID - 1;
         }
@@ -230,6 +251,9 @@ public class TestFileStore extends FileStoreImpl {
                 .forEach(
                         w -> {
                             try {
+                                // wait for compaction to end, otherwise orphan files may occur
+                                // see CompactManager#cancelCompaction for more info
+                                w.sync();
                                 w.close();
                             } catch (Exception e) {
                                 throw new RuntimeException(e);
@@ -238,7 +262,7 @@ public class TestFileStore extends FileStoreImpl {
 
         List<Snapshot> snapshots = new ArrayList<>();
         for (long id = snapshotIdBeforeCommit + 1; id <= snapshotIdAfterCommit; id++) {
-            snapshots.add(Snapshot.fromPath(pathFactory.toSnapshotPath(id)));
+            snapshots.add(snapshotManager.snapshot(id));
         }
         return snapshots;
     }
@@ -256,27 +280,29 @@ public class TestFileStore extends FileStoreImpl {
             }
         }
 
-        Map<BinaryRowData, Map<Integer, List<SstFileMeta>>> filesPerPartitionAndBucket =
+        Map<BinaryRowData, Map<Integer, List<DataFileMeta>>> filesPerPartitionAndBucket =
                 new HashMap<>();
         for (ManifestEntry entry : entries) {
             filesPerPartitionAndBucket
-                    .compute(entry.partition(), (p, m) -> m == null ? new HashMap<>() : m)
-                    .compute(entry.bucket(), (b, l) -> l == null ? new ArrayList<>() : l)
+                    .computeIfAbsent(entry.partition(), p -> new HashMap<>())
+                    .computeIfAbsent(entry.bucket(), b -> new ArrayList<>())
                     .add(entry.file());
         }
 
         List<KeyValue> kvs = new ArrayList<>();
-        FileStoreRead read = newRead();
-        for (Map.Entry<BinaryRowData, Map<Integer, List<SstFileMeta>>> entryWithPartition :
+        FileStoreRead<KeyValue> read = newRead();
+        for (Map.Entry<BinaryRowData, Map<Integer, List<DataFileMeta>>> entryWithPartition :
                 filesPerPartitionAndBucket.entrySet()) {
-            for (Map.Entry<Integer, List<SstFileMeta>> entryWithBucket :
+            for (Map.Entry<Integer, List<DataFileMeta>> entryWithBucket :
                     entryWithPartition.getValue().entrySet()) {
-                RecordReaderIterator iterator =
-                        new RecordReaderIterator(
+                RecordReaderIterator<KeyValue> iterator =
+                        new RecordReaderIterator<>(
                                 read.createReader(
-                                        entryWithPartition.getKey(),
-                                        entryWithBucket.getKey(),
-                                        entryWithBucket.getValue()));
+                                        new Split(
+                                                entryWithPartition.getKey(),
+                                                entryWithBucket.getKey(),
+                                                entryWithBucket.getValue(),
+                                                false)));
                 while (iterator.hasNext()) {
                     kvs.add(iterator.next().copy(keySerializer, valueSerializer));
                 }
@@ -286,14 +312,28 @@ public class TestFileStore extends FileStoreImpl {
     }
 
     public Map<BinaryRowData, BinaryRowData> toKvMap(List<KeyValue> kvs) {
+        if (LOG.isDebugEnabled()) {
+            LOG.debug(
+                    "Compacting list of key values to kv map\n"
+                            + kvs.stream()
+                                    .map(
+                                            kv ->
+                                                    kv.toString(
+                                                            TestKeyValueGenerator.KEY_TYPE,
+                                                            TestKeyValueGenerator.DEFAULT_ROW_TYPE))
+                                    .collect(Collectors.joining("\n")));
+        }
+
         Map<BinaryRowData, BinaryRowData> result = new HashMap<>();
         for (KeyValue kv : kvs) {
             BinaryRowData key = keySerializer.toBinaryRow(kv.key()).copy();
             BinaryRowData value = valueSerializer.toBinaryRow(kv.value()).copy();
             switch (kv.valueKind()) {
-                case ADD:
+                case INSERT:
+                case UPDATE_AFTER:
                     result.put(key, value);
                     break;
+                case UPDATE_BEFORE:
                 case DELETE:
                     result.remove(key);
                     break;
@@ -318,18 +358,19 @@ public class TestFileStore extends FileStoreImpl {
         // possibly not the most accurate, so this check is only.
         // - latest should < true_latest
         // - earliest should < true_earliest
-        Path snapshotDir = pathFactory().snapshotDirectory();
-        Path earliest = new Path(snapshotDir, SnapshotFinder.EARLIEST);
-        Path latest = new Path(snapshotDir, SnapshotFinder.LATEST);
+        SnapshotManager snapshotManager = snapshotManager();
+        Path snapshotDir = snapshotManager.snapshotDirectory();
+        Path earliest = new Path(snapshotDir, SnapshotManager.EARLIEST);
+        Path latest = new Path(snapshotDir, SnapshotManager.LATEST);
         if (actualFiles.remove(earliest)) {
-            long earliestId = SnapshotFinder.readHint(snapshotDir, SnapshotFinder.EARLIEST);
+            long earliestId = snapshotManager.readHint(SnapshotManager.EARLIEST);
             earliest.getFileSystem().delete(earliest, false);
-            assertThat(earliestId <= SnapshotFinder.findEarliest(snapshotDir)).isTrue();
+            assertThat(earliestId <= snapshotManager.findEarliest()).isTrue();
         }
         if (actualFiles.remove(latest)) {
-            long latestId = SnapshotFinder.readHint(snapshotDir, SnapshotFinder.LATEST);
+            long latestId = snapshotManager.readHint(SnapshotManager.LATEST);
             latest.getFileSystem().delete(latest, false);
-            assertThat(latestId <= SnapshotFinder.findLatest(snapshotDir)).isTrue();
+            assertThat(latestId <= snapshotManager.findLatest()).isTrue();
         }
         actualFiles.remove(latest);
 
@@ -340,30 +381,24 @@ public class TestFileStore extends FileStoreImpl {
         FileStorePathFactory pathFactory = pathFactory();
         ManifestList manifestList = manifestListFactory().create();
         FileStoreScan scan = newScan();
-        FileStorePathFactory.SstPathFactoryCache sstPathFactoryCache =
-                new FileStorePathFactory.SstPathFactoryCache(pathFactory);
 
-        Long latestSnapshotId = pathFactory.latestSnapshotId();
+        SnapshotManager snapshotManager = snapshotManager();
+        Long latestSnapshotId = snapshotManager.latestSnapshotId();
         if (latestSnapshotId == null) {
             return Collections.emptySet();
         }
 
         long firstInUseSnapshotId = Snapshot.FIRST_SNAPSHOT_ID;
         for (long id = latestSnapshotId - 1; id >= Snapshot.FIRST_SNAPSHOT_ID; id--) {
-            Path snapshotPath = pathFactory.toSnapshotPath(id);
-            try {
-                if (!snapshotPath.getFileSystem().exists(snapshotPath)) {
-                    firstInUseSnapshotId = id + 1;
-                    break;
-                }
-            } catch (IOException e) {
-                throw new RuntimeException(e);
+            if (!snapshotManager.snapshotExists(id)) {
+                firstInUseSnapshotId = id + 1;
+                break;
             }
         }
 
         Set<Path> result = new HashSet<>();
         for (long id = firstInUseSnapshotId; id <= latestSnapshotId; id++) {
-            Path snapshotPath = pathFactory.toSnapshotPath(id);
+            Path snapshotPath = snapshotManager.snapshotPath(id);
             Snapshot snapshot = Snapshot.fromPath(snapshotPath);
 
             // snapshot file
@@ -377,13 +412,13 @@ public class TestFileStore extends FileStoreImpl {
             List<ManifestFileMeta> manifests = snapshot.readAllManifests(manifestList);
             manifests.forEach(m -> result.add(pathFactory.toManifestFilePath(m.fileName())));
 
-            // sst
+            // data file
             List<ManifestEntry> entries = scan.withManifestList(manifests).plan().files();
             for (ManifestEntry entry : entries) {
                 result.add(
-                        sstPathFactoryCache
-                                .getSstPathFactory(entry.partition(), entry.bucket())
-                                .toPath(entry.file().fileName()));
+                        new Path(
+                                pathFactory.bucketPath(entry.partition(), entry.bucket()),
+                                entry.file().fileName()));
             }
         }
         return result;
